# -*- coding: utf-8 -*-
#
# log.py
#
# This file is part of py2048.
#
# py2048 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py2048 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py2048.  If not, see <https://www.gnu.org/licenses/>.

import sys
from pathlib import Path
import logging

from py2048 import APPNAME, DATA_DIR


__all__ = ["logger", "setup_logger"]


# setup log folder and log file path
DATA_DIR.mkdir(exist_ok=True)
log_file = str(DATA_DIR.resolve() / APPNAME / ".log")

# setup the default handlers
HANDLERS = (
    logging.FileHandler(filename=log_file, mode="w"),
    logging.StreamHandler(stream=sys.stderr),
)
HANDLERS[-1].setLevel(logging.ERROR)  # will override the logger's level

# set up a formatter (to be used for all handlers)
FORMATTER = logging.Formatter(
    fmt="\t".join(
        [
            # "%(asctime)s,%(msecs)d",  # human-readable timestamp
            "%(asctime)s",  # human-readable timestamp
            # "%(name)s @ %(module)s",  # logger name @ module
            "%(funcName)s @ %(module)s",  # calling function @ module
            "(%(levelname)s) %(message)s",  # (level name) message
        ]
    ),
    datefmt="%H:%M:%S",
)


def setup_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    for h in HANDLERS:
        h.setFormatter(FORMATTER)
        logger.addHandler(h)
        logger.info("Loaded handler: '%s'.", h)
    logger.info("Created '%s'.", logger)
    return logger


# set up the main logger and equip it
logger = setup_logger(APPNAME)



# -*- coding: utf-8 -*-
#
# cell.py
#
# This file is part of py2048.
#
# py2048 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py2048 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py2048.  If not, see <https://www.gnu.org/licenses/>.

from typing import Sequence
from functools import partialmethod

from py2048 import check_int, Point, type_check


class Cell:
    def __init__(self, point: Point, number: int = 0) -> None:
        type_check(point, Point)
        self.point = point
        self.x, self.y = point
        self._lock = False
        # don't use self._number here; we want bad initial values to be
        # detected
        self.number: int = number

    @property
    def number(self) -> int:
        return self._number

    @number.setter
    def number(self, value: int) -> None:
        check_int(value)
        self._number = value

    def __bool__(self) -> bool:
        return bool(self.number)

    def __eq__(self, other) -> bool:
        """Docs state that 'If a class does not define an `__eq__()` method it
        should not define a `__hash__()` operation either'. We want to define
        `__hash__`, so we must define `__eq__` as well. Two Cells are the same
        Cell iff they're at the same place, no matter their number.
        """

        if isinstance(other, type(self)):
            return self.point == other.point
        return NotImplemented

    def __hash__(self) -> int:
        """We want Cells to be hashable so that they can be used in sets.
        """

        # if two objects test as equal, then they MUST have the same hash value
        # objects that have a hash MUST produce the same hash over time
        return hash((self.__class__, self.x, self.y))

    def __gt__(self, other) -> bool:
        """Defined just to allow Cells to be ordered and sorted.
        """

        if isinstance(other, self.__class__):
            return self.point > other.point
        return NotImplemented

    def _locked_setter(self, onoff: bool) -> None:
        type_check(onoff, bool)
        if onoff and not self:
            # A 0-Cell shouldn't be moving in the first place, so it can't be
            # locked!
            raise ValueError("It doesn't make sense to lock an empty Cell")
        self._lock = onoff

    locked = property(
        fget=lambda self: self._lock,
        fset=_locked_setter,
        doc=(
            "The 'lock' is a `bool` that tells whether the `Cell` has already "
            "moved this cycle. This prevents a 2 merging into a 2 and the "
            "resulting 4 merging into another 4 all in a single movement."
        ),
    )

    lock = partialmethod(_locked_setter, True)
    unlock = partialmethod(_locked_setter, False)

    def __repr__(self) -> str:
        return f"Cell<{self.x},{self.y}>({self.number})"

    def __str__(self) -> str:
        return str(self.number)


Tissue = Sequence[Cell]



# -*- coding: utf-8 -*-
#
# frontends/basefrontend.py
#
# This file is part of py2048.
#
# py2048 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py2048 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py2048.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, final, List, Optional
from abc import ABC
import sys
from pathlib import Path
import random

from py2048 import Directions, type_check
from py2048.grid import Grid


class Base2048Frontend(ABC):
    """This abstract class implements the `play2048` method. A concrete
    frontend should inherit from this class, but shouldn't override this
    method.
    Many other methods are 'hooks' that either can, should, or must be
    overriden.
    """

    # 2^11 == 2048
    # sys.maxsize == (2^63) - 1 on 64bits machines
    valid_numbers: List[int] = [2 ** power for power in range(1, 12)]
    DIRECTIONS = tuple(Directions)
    SLEEP_S = 1  # how many seconds to sleep when in auto mode
    # converting seconds to mseconds
    SLEEP_MS = int(SLEEP_S * 1_000)

    @classmethod
    def is2048like(cls, n: int) -> bool:
        cache = cls.valid_numbers
        while cache[-1] < sys.maxsize and n > cache[-1]:
            cache.append(cache[-1] * 2)
        return n in cache

    def restart(self) -> None:
        self.victory = False
        self.grid.reset()
        self.grid.seed()

    def __init__(
        self,
        grid: Grid,
        is_random: bool,
        *,  # makes the remaining arguments keyword-only
        goal: Optional[int] = None,
    ) -> None:
        # first, perform some checks
        type_check(grid, Grid)
        # `goal = goal or 2048` would allow "goal = 0" to pass silently
        if goal is None:
            goal = 2048
        elif not self.is2048like(goal):
            raise ValueError(
                "goal must be a positive power of 2 "
                f"smaller than {sys.maxsize}"
            )
        # checks passed
        self.grid = grid
        self.goal = goal
        if is_random:
            self.choice_function = self.guess_direction
        else:
            self.choice_function = self.choose_direction
        self.is_random = is_random
        self.victory = False

    # play hooks: don't need to be overriden, but probably should
    def on_play(self) -> Any:
        pass

    def on_attempt(self) -> Any:
        pass

    def after_choice(self, choice: str) -> Any:
        pass

    def after_change(self, choice: str) -> Any:
        pass

    def after_nochange(self, choice: str) -> Any:
        pass

    def after_attempt(self, choice: str) -> Any:
        pass

    def after_play(self) -> Any:
        pass

    def guess_direction(self) -> Directions:
        return random.choice(self.DIRECTIONS)

    # MAIN LOOP: shouldn't be overriden
    @final
    def play2048(self) -> None:
        """The main loop repeatedly calls self.choice_function. If that raises
        `KeyboardInterrupt`, it returns immediately, without even calling
        `self.after_play()`. If that raises `EOFError`, it breaks the loop and
        calls `self.after_play()`.
        This loops until either a) the grid is jammed; or b) the player exited;
        or c) the goal has been reached for the first time.
        """

        player_quit = False
        grid = self.grid
        if grid.is_empty:  # can't play with an empty board
            grid.seed()
        assert not grid.is_empty
        self.on_play()
        # the actual loop
        jammed = grid.jammed
        assert not jammed
        while not jammed:
            self.on_attempt()
            try:
                choice = self.choice_function()
            except KeyboardInterrupt:
                # quickly exit due to Ctrl-C
                print()  # makes terminal looks nicer
                return
            except EOFError:
                # Ctrl-D/Z or some quit-command
                player_quit = True
                break
            assert choice in self.DIRECTIONS
            self.after_choice(choice)
            dragged = grid.drag(choice)
            if dragged:
                self.after_change(choice)
                jammed = grid.jammed
            else:
                self.after_nochange(choice)
            self.after_attempt(choice)
            if not self.victory and grid.largest_number >= self.goal:
                self.victory = True
                break
        # after loop stuff
        self.after_play()
        if player_quit:
            self.on_player_quit()
        elif self.victory:
            if jammed:
                # the game has jammed, but the player had already won, so it's
                # an "overvictory"
                self.on_player_overvictory()
            else:
                # player's winning for the first time
                self.on_player_victory()
        else:
            assert jammed
            self.on_player_loss()

    # the following methods MUST be overriden
    def choose_direction(self) -> Directions:
        raise NotImplementedError(
            "'choose_direction' must be overriden by subclass"
        )

    def on_player_quit(self) -> Any:
        raise NotImplementedError(
            "'on_player_quit' must be overriden by subclass"
        )

    def on_player_victory(self) -> Any:
        raise NotImplementedError(
            "'on_player_victory' must be overriden by subclass"
        )

    def on_player_overvictory(self) -> Any:
        raise NotImplementedError(
            "'on_player_overvictory' must be overriden by subclass"
        )

    def on_player_loss(self) -> Any:
        raise NotImplementedError(
            "'on_player_loss' must be overriden by subclass"
        )



# -*- coding: utf-8 -*-
#
# __init__.py
#
# This file is part of py2048.
#
# py2048 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py2048 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py2048.  If not, see <https://www.gnu.org/licenses/>.

# from __future__ import annotations
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    # Type,
    TYPE_CHECKING,
    Union,
)
from abc import ABC
import sys
from pathlib import Path
from collections import namedtuple
from collections.abc import Sequence as AbsSequence
from enum import auto, Enum, unique
from itertools import count, repeat

import appdirs  # https://pypi.org/project/appdirs
from more_itertools import unzip


## CONSTANTS
# generic constants
# https://github.com/python/cpython/blob/ebe20d9e7eb138c053958bc0a3058d34c6e1a679/Lib/types.py#L51
ModuleType = type(sys)  # just for annotation purposes
Vector = Sequence[int]
# https://github.com/python/typing/issues/684#issuecomment-548203158
if TYPE_CHECKING:

    class EllipsisType(Enum):
        Ellipsis = "..."

    Ellipsis = EllipsisType.Ellipsis
else:
    EllipsisType = type(Ellipsis)
EMPTY_TUPLE = tuple()  # dummy used in `min` and `max` calls
INFTY = float("inf")
NULL_SLICE = slice(None)
# specific constants
APPNAME = __name__
__version__ = (0, 13)
VERSION = ".".join(map(str, __version__))
TESTING = False
DATA_DIR = Path(appdirs.user_data_dir(appname=APPNAME))


__all__ = [
    # global variables
    "INFTY",
    # generic functions
    "iscontainer",
    "type_check",
    "check_int",
    # exceptions
    "Base2048Error",
    "NegativeIntegerError",
    "ExpectationError",
    # generic classes
    "Point",
    "Directions",
    "GridIndex",
    # game grid classes
    "BaseGameGrid",
    "SquareGameGrid",
]


## GENERAL-PURPOSE FUNCTIONS
def iscontainer(thing: Any) -> bool:
    """Determines whether the argument is iterable, but not a `str`.
    """

    cls = type(thing)
    return issubclass(cls, AbsSequence) and cls is not str


def type_check(
    value: Any, expected: Union[type, Sequence[type]], positive: bool = True
) -> None:
    """If the `positive` argument is True, raises an ExpectationError if the
    type of `value` is not listed in, or differs from, `expected`.
    If `positive` is False, raises the error if the type of `value` is listed
    in, or equals, `expected`.
    """

    if iscontainer(expected):
        match = type(value) in expected
    else:
        match = isinstance(value, expected)
    if match != positive:
        raise ExpectationError(value, expected, positive=positive)


def check_int(i: int) -> None:
    """Raises `ExpectationError` if `obj` is not `int`, and
    `NegativeIntegerError` if it's a negative `int`.
    """

    type_check(i, int)
    if i < 0:
        raise NegativeIntegerError(i)


## CLASSES
# Exceptions
class Base2048Error(Exception):
    pass


class NegativeIntegerError(Base2048Error, ValueError):
    """Raised when a negative `int` is found when a positive one or 0 was
    required.
    """

    std_message = f"Only non-negative integers can be used in a {APPNAME} grid"

    def __init__(self, number: int, message: str = "") -> None:
        self.number = number
        self.message = f"{message}. " if message else ""
        super().__init__(message)

    def __str__(self) -> str:
        return f"{self.message}{self.std_message}; but {self.number} found"


class ExpectationError(Base2048Error, TypeError):
    """Raised when the type of an argument is incorrect. Just like `TypeError`,
    but more verbose.
    """

    def __init__(
        self,
        problem: Any,
        expectations: Any,
        *args: str,
        positive: bool = True,
    ) -> None:
        if iscontainer(expectations):
            self.expectations = "/".join([e.__name__ for e in expectations])
        else:
            self.expectations = expectations.__name__
        self.positive = positive
        self.problem = repr(problem)
        self.problem_type = type(problem).__name__
        self.message = "" if not args else "".join(map(repr, args)) + ". "
        super().__init__(*args)

    def __str__(self) -> str:
        if self.positive:
            return (
                f"{self.message}Expected {self.expectations}, "
                f"but {self.problem} is {self.problem_type}"
            )
        return (
            f"{self.message}Found {self.problem} of type {self.problem_type}; "
            f"the following class(es) is/are not allowed: {self.expectations}"
        )


# other classes
class Point(namedtuple("Point", "x y")):
    """A namedtuple that stores a pair of non-negatives integers.
    """

    # `__slots__`is "a declaration inside a class that saves memory by
    # pre-declaring space for instance attributes and eliminating instance
    # dictionaries
    # https://docs.python.org/3/glossary.html#__slots__
    __slots__ = ()

    def __init__(self, x: int, y: int) -> None:
        check_int(x)
        check_int(y)

    def __repr__(self) -> str:
        return f"Point({self.x}, {self.y})"

    def __str__(self) -> str:
        return f"<{self.x},{self.y}>"


Line = Tuple[Point]


@unique
class Directions(Enum):
    """Enumerates the four orthogonal directions in the WASD order. The value
    of each direction is it's lowercased name: `Directions.UP.value == "up"`,
    and so on.
    """

    def _generate_next_value_(name, start, count, last_values) -> str:
        return name.lower()

    # WASD order
    UP = auto()
    LEFT = auto()
    DOWN = auto()
    RIGHT = auto()

    @classmethod
    def pretty(cls) -> str:
        return ", ".join(map(str, cls))


class GridIndex:
    """This class parses objects passed as indexes to a game grid (any instance
    of any subclass of `BaseGameGrid`.
    We want to allow those grids to be indexed in both directions in one
    statement.
    For example, grid[..., 3] should return the second row, grid[2, 3] should
    return the cell at the first column and the second row, and so on.
    grid[..., ...] copies the grid, just like L[:] copies a `list'.

    The class expects either a single argument, or a tuple with exactly two
    arguments.
    Each argument must be either an `Ellipsis' literal, an `int', or a `slice'.
    Integers cannot be negative.
    `Ellipsis' is converted into the "null slice", which is `slice(None)'. It
    represents that all rows, or all columns, are being selected.
    If only one object is passed, it is assigned to the x-axis, and the y-index
    defaults to the null slice.
    """

    SingletonType = Union[int, slice, EllipsisType]
    Type = Union[SingletonType, Tuple[SingletonType]]

    def __init__(self, *args: Sequence[SingletonType]) -> None:
        cls = type(self).__name__
        if not args:
            x = y = NULL_SLICE
        elif len(args) == 1:
            the_arg = args[0]
            if the_arg is None:
                raise TypeError(f"Cannot make a {cls} from None")
            # suppose it's a pair and try to unpack it
            try:
                x, y = the_arg
            except (ValueError, TypeError):
                x, y = the_arg, NULL_SLICE
        elif len(args) == 2:
            if None in args:
                raise TypeError(f"Cannot make a {cls} from None")
            x, y = args
        else:
            raise ValueError(
                f"Cannot create a {cls} with more than 2 arguments"
            )
        if x is Ellipsis:
            x = NULL_SLICE
        if y is Ellipsis:
            y = NULL_SLICE
        pair = (x, y)
        for i in pair:
            if not isinstance(i, slice):
                check_int(i)
        self.x = x
        self.y = y
        self.xy = pair
        self._str = str(pair)

    def __iter__(self) -> Iterator[SingletonType]:
        return iter(self.xy)

    def __getitem__(self, key: int) -> SingletonType:
        return self.xy[key]

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self.xy == other.xy
        return NotImplemented

    def __hash__(self) -> int:
        return hash((type(self), self.x, self.y))

    def __repr__(self) -> str:
        return type(self).__name__ + self._str

    def __str__(self) -> str:
        return self._str


## MAIN CLASS IN THIS MODULE
class BaseGameGrid(ABC):
    """Wrapper over `dict` that maps Points into objects that must have the
    type CELLCLASS. CELLCLASS is `None`, which means that it must be overidden
    when subclassing this class.
    The class can abstract any table with orthogonally aligned square cells,
    such as the chess and checkers boards, sudoku grids, and so on.
    While the cells must be squares, the table itself can be a rectangle.
    The class provides basic functionality such as iterating over its rows and
    columns and can be indexed by a `GridIndex`, as explained in the docstring
    of that class.
    """

    CELLCLASS: Any = None

    def __init__(self, width: int, height: int) -> None:
        cls = self.CELLCLASS
        if cls is None:
            raise NotImplementedError(
                "CELLCLASS must be overriden when subclassing BaseGameGrid"
            )
        check_int(width)
        check_int(height)
        self.map: Dict[Point, cls] = {}
        for x in range(width):
            for y in range(height):
                point = Point(x, y)
                try:
                    self.map[point] = cls(point)
                except TypeError:
                    self.map[point] = cls()

    def keys(self) -> Iterator[Point]:
        return iter(sorted(self.map.keys()))

    def values(self) -> Iterator[CELLCLASS]:
        return iter((self.map[k] for k in self.keys()))

    def items(self) -> Iterator[Tuple[Point, CELLCLASS]]:
        return iter(((k, self.map[k]) for k in self.keys()))

    # some synonyms
    __iter__ = keys
    points = keys

    def __len__(self) -> int:
        # this also provides __bool__
        return len(self.map)

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self.map == other.map
        return NotImplemented

    @property
    def x_axis(self) -> Iterator[int]:
        yielded = set()
        for x in unzip(self.keys())[0]:
            if x not in yielded:
                yield x
                yielded.add(x)

    @property
    def y_axis(self) -> Iterator[int]:
        yielded = set()
        for y in unzip(self.keys())[1]:
            if y not in yielded:
                yield y
                yielded.add(y)

    @property
    def height(self) -> int:
        return len(tuple(self.y_axis))

    @property
    def width(self) -> int:
        return len(tuple(self.x_axis))

    @property
    def rows(self) -> Iterator[Line]:
        for y in self.y_axis:
            yield self[..., y]

    @property
    def columns(self) -> Iterator[Line]:
        for x in self.x_axis:
            yield self[x]

    def check_integrity(self) -> None:
        # check columns
        # the following ensures that there are no "gaps" nor "jumps" in the
        # column indexes, and then in the row indexes
        for actual, target in zip(self.x_axis, count()):
            if actual != target:
                raise ValueError(
                    f"Wrong column index {actual}; must be {target}"
                )
        shortest = min(tuple(self.columns), default=EMPTY_TUPLE)
        longest = max(tuple(self.columns), default=EMPTY_TUPLE)
        assert len(shortest) == len(longest)
        # check rows
        for actual, target in zip(self.y_axis, count()):
            if actual != target:
                raise ValueError(f"Wrong row index {actual}; must be {target}")
        shortest = min(tuple(self.rows), default=EMPTY_TUPLE)
        longest = max(tuple(self.rows), default=EMPTY_TUPLE)
        assert len(shortest) == len(longest)
        # type check and total size check
        for value in self.values():
            assert isinstance(value, self.CELLCLASS)
        assert len(self) == self.height * self.width

    @staticmethod
    def slice2range(s: slice) -> range:
        # slice objects have read-only data attributes start, stop and step
        # which merely return the argument values (or their default).
        start = s.start or 0
        stop = s.stop or sys.maxsize
        step = s.step or 1
        return range(start, stop, step)

    def select_keys(self, key: GridIndex.Type) -> Line:
        x, y = GridIndex(key)
        restrictions: Set[Callable] = set()
        # restricting the X-axis
        if isinstance(x, int):
            restrictions.add(lambda key: key.x == x)
        elif isinstance(x, slice) and x != NULL_SLICE:
            x_range = self.slice2range(x)
            restrictions.add(lambda key: key.x in x_range)
        # restricting the Y-axis
        if isinstance(y, int):
            restrictions.add(lambda key: key.y == y)
        elif isinstance(y, slice) and y != NULL_SLICE:
            y_range = self.slice2range(y)
            restrictions.add(lambda key: key.y in y_range)
        # assembling everything
        selected: List[Point] = []
        for this_key in self:
            for restriction in restrictions:
                # a break in a for-loop skips its else-clause
                if not restriction(this_key):
                    # the first restriction that doesn't apply is enough to
                    # discard this key
                    break
            else:  # happens iff the for-clause didn't end with a break
                selected.append(this_key)
        return tuple(selected)

    def __getitem__(self, key: GridIndex.Type) -> Union[Point, Line]:
        if isinstance(key, Point):
            # no need for further complications when getting a single Point
            return self.map[key]
        selected = tuple(
            (self.map[selected] for selected in self.select_keys(key))
        )
        if not selected:
            raise KeyError
        if len(selected) == 1:
            return selected[0]
        return selected

    def set_point(self, key: Point, value: Any) -> None:
        """This method exists to be overriden by subclasses, if necessary.
        """

        type_check(value, self.CELLCLASS)
        self.map[key] = value

    def __setitem__(
        self, key: GridIndex.Type, value: Union[int, CELLCLASS]
    ) -> None:
        if isinstance(key, Point):
            # no need for further complications when setting a single Point
            self.set_point(key, value)
            return
        selected = self.select_keys(key)
        amount = len(selected)
        if not iscontainer(value):
            value = repeat(value, amount)
        elif len(value) != amount:
            raise ValueError(
                "Wrong amount of values to unpack: "
                f"expected {amount}, got {len(value)}"
            )
        for point, n in zip(selected, value):
            self.set_point(point, n)
        self.check_integrity()

    def __repr__(self) -> str:
        cls = type(self).__name__
        cellclass = self.CELLCLASS.__name__
        w, h = self.width, self.height
        return f"{cls}('{cellclass}', {w} x {h})"

    def __str__(self) -> str:
        return "\n".join(map(repr, tuple(self.rows)))


class SquareGameGrid(BaseGameGrid):
    """Exactly like BaseGameGrid, but enforced to remain a square table.
    """

    def __init__(self, side: int) -> None:
        super().__init__(side, side)

    def check_integrity(self) -> None:
        super().check_integrity()
        if self.width != self.height:
            raise ValueError(
                "A SquareGameGrid cannot have different side lenghts, but "
                f"width is {self.width} and height is {self.height}"
            )


def demonstration() -> None:
    # from random import randint

    class IntGameGrid(BaseGameGrid):
        CELLCLASS = int

    g = IntGameGrid(4, 3)
    print(g)
    print("-" * 79)

    for col in range(g.width):
        g[col, ...] = col
    g[..., 0] = (9, 8, 7, 6)

    print(g)
    print("-" * 79)

    rows, cols = tuple(g.rows), tuple(g.columns)
    print(rows)
    print(cols)
    print(g[1:, 1:])


if __name__ == "__main__":
    demonstration()



# -*- coding: utf-8 -*-
#
# setup.py

import setuptools

from py2048 import APPNAME, TESTING, VERSION


def read(name: str) -> None:
    with open(name, "r", encoding="utf-8") as readme:
        return readme.read()


setuptools.setup(
    name=f"{APPNAME}-danieldiniz" if TESTING else f"{APPNAME}base",
    version=VERSION,
    author="Daniel Diniz",
    author_email="daniel_asl_diniz@protonmail.com",
    description="Python clone of the famous 2048 game.",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    url = "https://github.com/dasld/py2048base",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3.8",
        "Topic :: Games/Entertainment :: Puzzle Games",
    ],
    install_requires=[  # alphabetical order
        "appdirs>=1",
        "more-itertools>=8",
    ],
    packages=[APPNAME],
    python_requires=">=3.8",
    license="GPL",
    zip_safe=True,
)



# -*- coding: utf-8 -*-
#
# grid.py
#
# This file is part of py2048.
#
# py2048 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py2048 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py2048.  If not, see <https://www.gnu.org/licenses/>.

from typing import (
    # Any,
    Optional,
    Set,
    Tuple,
)
import logging
import random
from itertools import chain

from py2048 import (
    SquareGameGrid,
    Directions,
    ExpectationError,
    NegativeIntegerError,
    Point,
)
from py2048.cell import Cell


logger = logging.getLogger(__name__)


class Grid(SquareGameGrid):
    CELLCLASS: type = Cell
    # how many cells start non-zero
    STARTING_AMOUNT = 2
    # what values can be seeded in each cell
    SEEDING_VALUES = (2, 4)

    cells = SquareGameGrid.values

    def __init__(self, side: int = 4) -> None:
        logger.info("Creating a %dx%d Grid instance", side, side)
        assert 0 < self.STARTING_AMOUNT < side ** 2
        super().__init__(side)
        self.empty_cells: Set[Cell] = set(self.cells())
        self.reset(on_init=True)

    def reset(self, on_init: bool = False) -> None:
        self.attempt = 0
        self.cycle = 0
        self.score = 0
        if not on_init:
            for cell in self.cells():
                if cell:
                    cell.unlock()
                    self.set_cell(cell, 0)
        self.check_integrity()

    @property
    def largest_number(self) -> int:
        return max([cell.number for cell in self.cells()], default=0)

    @property
    def is_empty(self) -> bool:
        """Returns whether every Cell is empty (ie, 0-numbered).
        """
        assert len(self.empty_cells) <= len(self)
        return len(self.empty_cells) == len(self)

    def set_cell(self, cell: Cell, number: int) -> None:
        """Assigns a number to a Cell and updates `empty_cells`.
        """

        cell.number = number
        if number and (cell in self.empty_cells):
            self.empty_cells.remove(cell)
        elif not number and (cell not in self.empty_cells):
            self.empty_cells.add(cell)

    def set_point(self, key: Point, number: int) -> None:
        """Overrides a method of the base class. To assign a number to a Point
        is to change its Cell's number, not to replace the Cell with the
        number.
        This method is used by `BaseGameGrid` methods such as `__setitem__`.
        """

        self.set_cell(self[key], number)

    def seed(self, amount: int = STARTING_AMOUNT) -> None:
        """Assigns a number randomly chosen from `Grid.SEEDING_VALUES` to a
        few (`amount`) randomly selected empty Cells.
        """

        logger.debug(
            "Available cells for seeding: %s.",
            "  ".join(sorted(map(repr, self.empty_cells))),
        )
        changed: Set[Cell] = set()
        for cell in random.sample(self.empty_cells, amount):
            assert not cell  # must be 0
            self.set_cell(cell, random.choice(self.SEEDING_VALUES))
            changed.add(cell)
        logger.debug(
            "Seeded those cells: %s.", "  ".join(sorted(map(repr, changed))),
        )
        # return tuple(changed)

    def neighbor(self, cell: Cell, to: Directions) -> Optional[Cell]:
        """Returns the next (possibly empty) Cell in the given direction, or
        `None` if there"s none (if `self` is at the edge of the board,
        for example).
        """

        x, y = cell.point
        if to == Directions.LEFT:
            x -= 1
        elif to == Directions.RIGHT:
            x += 1
        elif to == Directions.DOWN:
            y += 1
        elif to == Directions.UP:
            y -= 1
        else:
            raise ValueError(f"'to' must be {Directions.pretty()}, not {to!r}")
        try:
            next_point = Point(x, y)  # NegativeIntegerError
            next_cell = self[next_point]  # KeyError
        except (NegativeIntegerError, KeyError):
            return None
        assert isinstance(next_cell, type(cell))
        return next_cell

    def pivot(self, cell: Cell, to: Directions) -> Cell:
        """Returns the first non-empty, unlocked Cell in the given direction,
        including itself. If all Cells in this direction are empty, returns the
        one at the edge of the board instead.
        """

        current = last = cell
        while True:
            current = self.neighbor(current, to)
            if current is None:  # edge of the board; give up
                return last
            if not current:  # empty Cell; keep looking
                last = current
                continue
            if current.number == cell.number and not current.locked:
                return current
            # in this case, the new Cell's number matches, but it is locked
            # because it has already moved this cycle, so we ignore it
            return last

    def move_cell(self, cell: Cell, to: Directions) -> bool:
        """Attempts to move `cell` in the given `direction`. When successful,
        locks the pivot and updates the score (if the starting number was
        positive), then updates the numbers of the `cell` and its pivot.
        Returns whether movement ocurred.
        """

        if not cell:  # empty Cells don't move
            return False
        pivot = self.pivot(cell, to)
        if pivot == cell:
            return False
        new_number = cell.number + pivot.number
        if pivot:
            pivot.lock()  # prevent further movement this cycle
            # update the score only when the Cell moved to a positive pivot
            self.score += new_number
        self.set_cell(cell, 0)
        self.set_cell(pivot, new_number)
        return True

    def drag(self, to: Directions) -> bool:
        """Increments its attempts count, then tries to move every cell in the
        given direction, starting with the vectors closest to this direction.
        If anything changed, it increments its cycle by 1 and seeds itself by
        1.
        """

        if to == Directions.LEFT:
            vectors = tuple(self.columns)
        elif to == Directions.RIGHT:
            vectors = tuple(self.columns)[::-1]
        elif to == Directions.UP:
            vectors = tuple(self.rows)
        elif to == Directions.DOWN:
            vectors = tuple(self.rows)[::-1]
        elif isinstance(to, Directions):
            # very bad
            raise ValueError(f"'to' must be {Directions.pretty()}, not {to!r}")
        else:
            # EVEN WORSE
            raise ExpectationError(to, Directions)
        something_moved = False
        self.attempt += 1
        logger.debug("Attempt increased to %d.", self.attempt)
        for cell in chain.from_iterable(vectors):
            this_moved = self.move_cell(cell, to)
            # assert isinstance(moved, bool)
            if not something_moved and this_moved:
                something_moved = True
        if something_moved:
            logger.debug("Dragging %r changed me.", to)
            self.cycle += 1
            logger.debug("Cycle increased to %d.", self.cycle)
            self.seed(1)
            for cell in self.cells():
                if cell:
                    cell.unlock()
        return something_moved

    @property
    def jammed(self) -> bool:
        """Returns False if there's any legal movement left for the player,
        True if there's none.
        Rationale: if there's at least one zero Cell, the Grid is not jammed,
        because the player can at least fill the empty spot.
        If there are no zeroes, check if any two adjacent Cells have the same
        number; stop as soon as any such pair is found.
        Is there a better/faster way of doing this without checking every Cell?
        """

        if empties := len(self.empty_cells):
            logger.debug("Not jammed: still %d empty cell(s).", empties)
            return False
        for cell in self.cells():
            for to in Directions:
                neighbor = self.neighbor(cell, to)
                if neighbor is None:
                    continue
                if cell.number == neighbor.number:
                    logger.debug(
                        "Not jammed: %r is next to %r.", cell, neighbor,
                    )
                    return False
        logger.debug("JAMMED!")
        return True
